#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from abc import ABC
import bisect
from typing import Callable
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import TypeVar

from dataloader.raw_dataset import RawDataset
from dataloader.transform import apply_transform
from dataloader.util.data_kind import DataKind

__all__ = ['Dataset', 'IterableDataset', 'BaseDataset', 'ChainDataset', 'ConcatDataset']

T_co = TypeVar('T_co', covariant=True)
T = TypeVar('T')


class BaseDataset(Generic[T_co]):
    """an abstract class representing a Dataset"""

    def __getitem__(self, index: int) -> T_co:
        raise NotImplementedError

    def __add__(self, other: 'BaseDataset[T_co]') -> 'ConcatDataset[T_co]':
        return ConcatDataset([self, other])


class _BaseIterableDataset(BaseDataset[T_co]):
    """an iterable Dataset"""
    def __iter__(self) -> Iterator[T_co]:
        raise NotImplementedError

    def __add__(self, other: BaseDataset[T_co]):
        return ChainDataset([self, other])

    def __getitem__(self, index: int):
        pass


class ConcatDataset(BaseDataset[T_co]):
    """assemble different existing datasets"""
    datasets: List[BaseDataset[T_co]]
    cumulative_sizes: List[int]

    @staticmethod
    def cumsum(sequence):
        r, s = [], 0
        for e in sequence:
            n = len(e)
            r.append(n + s)
            s += n

        return r

    def __init__(self, datasets: Iterable[BaseDataset]):
        super().__init__()
        self.datasets = list(datasets)

        self.cumulative_sizes = self.cumsum(self.datasets)

    def __len__(self):
        return self.cumulative_sizes[-1]

    def __getitem__(self, idx):
        if idx < 0:
            if -idx > len(self):
                raise ValueError('absolute value of index should not exceed dataset length')
            idx = len(self) + idx

        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
        if dataset_idx == 0:
            sample_idx = idx
        else:
            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]

        return self.datasets[dataset_idx][sample_idx]


class ChainDataset(_BaseIterableDataset):
    """assemble different existing dataset streams on-the-fly"""
    def __init__(self, datasets: Iterable[BaseDataset]):
        super().__init__()
        self.datasets = datasets

    def __iter__(self):
        for d in self.datasets:
            for x in d:
                yield x

    def __len__(self):
        ns = [len(d) for d in self.datasets]
        return sum(ns)


class IterableDataset(_BaseIterableDataset):
    def __init__(self, data: Iterable, transform: Optional[Callable] = None):
        self.data = data
        self.transform = transform

    def __iter__(self):
        data = iter(self.data)

        for d in data:
            d = apply_transform(self.transform, d)
            yield d


class Dataset(BaseDataset, RawDataset, ABC):
    def __init__(self, data, kind=DataKind.MEM_SEQ, transform: Optional[Callable] = None):
        """

        Args:
            data:
              - list of things: if kind=DataKind.MEM_SEQ
              - filename: if kind=DataKind.FILE or kind=DataKind.MMAP_FILE
            kind: @see DataKind
            transform: func for mapping raw data into feature
        """
        super().__init__(data, kind, transform)

    def __getitem__(self, index):
        return self.get(index)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import errno
import itertools
import queue
import tempfile
from typing import Any
from typing import Callable
from typing import Generic
from typing import List
from typing import Optional
from typing import Sequence
from typing import TypeVar

import multiprocess as multiprocessing  # when use multiprocessing, file descriptor cannot be pickled
import numpy as np
import tensorflow as tf

from dataloader import logger
from dataloader import util
from dataloader.dataset import BaseDataset
from dataloader.dataset import Dataset
from dataloader.dataset import IterableDataset
from dataloader.sampler import BatchSampler
from dataloader.sampler import InfiniteConstantSampler
from dataloader.sampler import RandomSampler
from dataloader.sampler import Sampler
from dataloader.sampler import SequentialSampler
from dataloader.util import signal_handling
from dataloader.util import to_tensor
from dataloader.util import worker
from dataloader.util.batch import RepeatInBatch
from dataloader.util.dataset_kind import DatasetKind
from dataloader.util.misc import ExceptionWrapper
from dataloader.util.misc import set_rnd

__all__ = ['DataLoader']


T_co = TypeVar('T_co', covariant=True)
T = TypeVar('T')

_fn_worker_init_t = Callable[[int], None]
_fn_to_tensor_t = Callable[[List[T]], Any]


class _BaseDataLoader(Generic[T_co]):
    """Data loader. Combines a dataset and a sampler, and provides an iterable over the given dataset"""

    dataset: BaseDataset[T_co]
    batch_size: Optional[int]
    num_workers: int
    drop_last: bool
    timeout: float
    sampler: Sampler
    prefetch_factor: int
    _iterator: Optional['_BaseDataLoaderIter']
    __initialized = False

    def __init__(
        self,
        dataset: BaseDataset[T_co],
        batch_size: Optional[int] = 1,
        shuffle: bool = False,
        sampler: Optional[Sampler[int]] = None,
        batch_sampler: Optional[Sampler[Sequence[int]]] = None,
        num_workers: int = 0,
        fn_to_tensor: _fn_to_tensor_t = None,
        drop_last: bool = False,
        timeout: float = 0,
        fn_worker_init: _fn_worker_init_t = None,
        multiprocessing_context=None,
        prefetch_factor: int = 2,
        persistent_workers: bool = False,
        repeat_in_batch: RepeatInBatch = None,
        transform: Optional[Callable] = None
    ):
        self.dataset = dataset
        self.num_workers = num_workers
        self.prefetch_factor = prefetch_factor
        self.timeout = timeout
        self.fn_worker_init = fn_worker_init
        self.multiprocessing_context = multiprocessing_context

        if isinstance(dataset, IterableDataset):
            self.dataset_kind = DatasetKind.Iterable
        else:
            self.dataset_kind = DatasetKind.Map

        if sampler is None:
            if self.dataset_kind == DatasetKind.Iterable:
                sampler = InfiniteConstantSampler()
            else:
                if shuffle:
                    sampler = RandomSampler(dataset)
                else:
                    sampler = SequentialSampler(dataset)

        if batch_size is not None and batch_sampler is None:
            batch_sampler = BatchSampler(sampler, batch_size, drop_last)

        self.batch_size = batch_size
        self.drop_last = drop_last
        self.sampler = sampler
        self.batch_sampler = batch_sampler

        if fn_to_tensor is None:
            fn_to_tensor = to_tensor.to_tf_tensor

        self.fn_to_tensor = fn_to_tensor
        self.persistent_workers = persistent_workers

        self.__initialized = True
        self._IterableDataset_len_called = None

        self._iterator = None

        self.repeat_in_batch = repeat_in_batch
        self.transform = transform

    def _get_iterator(self) -> '_BaseDataLoaderIter':
        if self.num_workers == 0:
            return _SingleProcessDataLoaderIter(self)

        return _MultiProcessingDataLoaderIter(self)

    @property
    def multiprocessing_context(self):
        return self.__multiprocessing_context

    @multiprocessing_context.setter
    def multiprocessing_context(self, multiprocessing_context):
        if multiprocessing_context is not None:
            if self.num_workers > 0:
                if isinstance(multiprocessing_context, (str, bytes)):
                    multiprocessing_context = multiprocessing.get_context(multiprocessing_context)

        self.__multiprocessing_context = multiprocessing_context

    def __setattr__(self, attr, val):
        if self.__initialized and attr in ('batch_size', 'sampler', 'drop_last', 'dataset', 'persistent_workers'):
            raise ValueError(
                '{} attribute should not be set after {} is initialized'.format(attr, self.__class__.__name__)
            )

        super().__setattr__(attr, val)

    def __iter__(self) -> '_BaseDataLoaderIter':
        if self.persistent_workers and self.num_workers > 0:
            if self._iterator is None:
                self._iterator = self._get_iterator()
            else:
                self._iterator.reset(self)

            return self._iterator

        return self._get_iterator()

    @property
    def is_batch(self):
        return self.batch_sampler is not None

    @property
    def index_sampler(self):
        if self.is_batch:
            return self.batch_sampler

        return self.sampler

    def __len__(self) -> int:
        if self.dataset_kind == DatasetKind.Iterable:
            length = self._IterableDataset_len_called = len(self.dataset)
            if self.batch_size is not None:
                from math import ceil
                if self.drop_last:
                    length = length // self.batch_size
                else:
                    length = ceil(length / self.batch_size)
            return length

        return len(self.index_sampler)


class _BaseDataLoaderIter:
    def __init__(self, loader: _BaseDataLoader) -> None:
        self._dataset = loader.dataset
        self._dataset_kind = loader.dataset_kind
        self._IterableDataset_len_called = loader._IterableDataset_len_called
        self._is_batch = loader.is_batch
        self._drop_last = loader.drop_last
        self._index_sampler = loader.index_sampler
        self._num_workers = loader.num_workers
        self._prefetch_factor = loader.prefetch_factor
        self._timeout = loader.timeout
        self._fn_to_tensor = loader.fn_to_tensor
        self._sampler_iter = iter(self._index_sampler)
        self._base_seed = tf.random.uniform((1, 1), 0, np.iinfo(np.int64).max, tf.int64).numpy()[0][0]
        self._persistent_workers = loader.persistent_workers
        self._num_yielded = 0

        self._repeat = loader.repeat_in_batch

    def __iter__(self) -> '_BaseDataLoaderIter':
        return self

    def reset(self, loader, first_iter=False):
        self._sampler_iter = iter(self._index_sampler)
        self._num_yielded = 0
        self._IterableDataset_len_called = loader._IterableDataset_len_called

    def _next_index(self):
        return next(self._sampler_iter)

    def _next_data(self):
        raise NotImplementedError

    def __next__(self) -> Any:
        data = self._next_data()

        self._num_yielded += 1

        return data

    def __len__(self) -> int:
        return len(self._index_sampler)


class _SingleProcessDataLoaderIter(_BaseDataLoaderIter):
    def __init__(self, loader):
        super().__init__(loader)

        self._dataset_fetcher = DatasetKind.create_fetcher(
            self._dataset_kind,
            self._dataset, self._is_batch, self._drop_last, self._fn_to_tensor, loader.transform, loader.repeat_in_batch
        )

    def _next_data(self):
        index = self._next_index()
        data = self._dataset_fetcher.fetch(index)

        return data


class _MultiProcessingDataLoaderIter(_BaseDataLoaderIter):
    """Iterates once over the DataLoader's dataset, as specified by the sampler"""
    def __init__(self, loader):
        super().__init__(loader)

        if loader.multiprocessing_context is None:
            multiprocessing_context = multiprocessing
        else:
            multiprocessing_context = loader.multiprocessing_context

        # DO NOT SET METHOD, it'll stuck the creation of workers
        # forkserver stuck
        # spawn: is killed by signal: Segmentation fault
        # fork:
        # dataset 中使用了 multiprocessing.Lock() 操作, 故需要与之保持一致, 使用 multiprocess as multiprocessing
        # 若不是一致, 则会报 RuntimeError: Lock objects should only be shared between processes through inheritance
        # 同时, 没有出现「创建 worker 时的 stuck」
        #
        # set multiprocessing method explicitly with spawn or forkserver
        # if not set: RuntimeError: DataLoader worker (pid(s)) exited unexpectedly
        # ref: https://github.com/horovod/horovod/issues/2053
        #      https://stackoverflow.com/questions/64095876/multiprocessing-fork-vs-spawn
        #      - fork: fork is fast, unsafe, and maybe bloated
        #      - spawn: spawn is safe, compact, and slower
        #      - forkserver: forkserver is fast, compact, and safe, but it's more complicated
        # try:
        #     multiprocessing_context.set_start_method(os.getenv('multiprocessing_context_method', 'forkserver'))
        # except RuntimeError as e:
        #     logger.warning(f'multiprocessing_context method has been set: {e}')
        #     pass

        self._fn_worker_init = loader.fn_worker_init
        self._worker_queue_idx_cycle = itertools.cycle(range(self._num_workers))

        self._worker_result_queue = multiprocessing_context.Queue()
        self._worker_pids_set = False
        self._shutdown = False
        self._workers_done_event = multiprocessing_context.Event()

        self._index_queues = []
        self._workers = []
        for worker_id in range(self._num_workers):
            index_queue = multiprocessing_context.Queue()
            index_queue.cancel_join_thread()

            w = multiprocessing_context.Process(
                target=worker.worker_loop,
                args=(
                    self._dataset_kind,
                    self._dataset, self._is_batch, self._drop_last, self._fn_to_tensor,
                    loader.transform, loader.repeat_in_batch,
                    index_queue, self._worker_result_queue, self._workers_done_event,
                    self._base_seed, self._fn_worker_init, worker_id, self._num_workers,
                )
            )

            w.daemon = True
            w.start()

            self._index_queues.append(index_queue)
            self._workers.append(w)

        self._data_queue = self._worker_result_queue

        self._worker_pids = {id(self): [w.pid for w in self._workers]}
        self._worker_pids_set = True

        signal_handling.set_sigchld_handler()

        self._send_idx = None
        self._rcvd_idx = None
        self._task_info = None
        self._tasks_outstanding = None
        self._workers_status = None

        self.reset(loader, first_iter=True)

    def reset(self, loader, first_iter=False):
        super().reset(loader, first_iter)

        self._send_idx = 0  # idx of the next task to be sent to workers
        self._rcvd_idx = 0  # idx of the next task to be returned in __next__

        # information about data not yet yielded, i.e., tasks w/ indices in range [rcvd_idx, send_idx).
        # map: task idx => - (worker_id,)        if data isn't fetched (outstanding)
        #                  \ (worker_id, data)   if data is already fetched (out-of-order)
        self._task_info = {}
        self._tasks_outstanding = 0  # always equal to count(v for v in task_info.values() if len(v) == 1)

        self._workers_status = [True for _ in range(self._num_workers)]

        # We resume the prefetching in case it was enabled
        if not first_iter:
            for idx in range(self._num_workers):
                self._index_queues[idx].put(worker.ResumeIteration())

            resume_iteration_cnt = self._num_workers
            while resume_iteration_cnt > 0:
                data = self._get_data()
                if isinstance(data, worker.ResumeIteration):
                    resume_iteration_cnt -= 1

        # prime the prefetch loop
        logger.info(f'prefetch {self._prefetch_factor * self._num_workers}')
        for _ in range(self._prefetch_factor * self._num_workers):
            self._try_put_index()

    def _try_get_data(self, timeout=worker.MP_STATUS_CHECK_INTERVAL):
        try:
            data = self._data_queue.get(timeout=timeout)
            return True, data
        except Exception as e:
            failed_workers = []
            for worker_id, w in enumerate(self._workers):
                if self._workers_status[worker_id] and not w.is_alive():
                    failed_workers.append(w)
                    self._mark_worker_as_unavailable(worker_id)

            if len(failed_workers) > 0:
                pids_str = ', '.join(str(w.pid) for w in failed_workers)
                raise RuntimeError(f'DataLoader worker (pid(s) {pids_str}) exited unexpectedly: {e}')

            if isinstance(e, queue.Empty):
                return False, None

            try:
                fds_limit_margin = 10
                _ = [tempfile.NamedTemporaryFile() for _ in range(fds_limit_margin)]
            except OSError as e:
                if e.errno == errno.EMFILE:
                    raise RuntimeError(
                        """Too many open files. Communication with the workers is no longer possible. 
                        Please increase the limit using `ulimit -n` in the shell
                        """
                    )
            raise

    def _get_data(self):
        if self._timeout > 0:
            success, data = self._try_get_data(self._timeout)
            if success:
                return data

            raise RuntimeError(f'DataLoader timed out after {self._timeout} seconds')
        else:
            while True:
                success, data = self._try_get_data()
                if success:
                    return data

    def _next_data(self):
        while True:
            while self._rcvd_idx < self._send_idx:
                info = self._task_info[self._rcvd_idx]
                if len(info) == 2 or self._workers_status[info[0]]:  # has data or is still active
                    break

                del self._task_info[self._rcvd_idx]
                self._rcvd_idx += 1
            else:
                # no valid `self._rcvd_idx` is found (i.e., didn't break)
                if not self._persistent_workers:
                    self._shutdown_workers()

                raise StopIteration

            # Now `self._rcvd_idx` is the batch index we want to fetch

            # Check if the next sample has already been generated
            if len(self._task_info[self._rcvd_idx]) == 2:
                data = self._task_info.pop(self._rcvd_idx)[1]
                return self._process_data(data)

            idx, data = self._get_data()
            self._tasks_outstanding -= 1
            if self._dataset_kind == DatasetKind.Iterable:
                if isinstance(data, worker.IterableDatasetStopIteration):
                    if self._persistent_workers:
                        self._workers_status[data.worker_id] = False
                    else:
                        self._mark_worker_as_unavailable(data.worker_id)

                    self._try_put_index()

                    continue

            if idx != self._rcvd_idx:
                # store out-of-order samples
                self._task_info[idx] += (data,)
            else:
                del self._task_info[idx]
                return self._process_data(data)

    def _try_put_index(self):
        try:
            index = self._next_index()
        except StopIteration:
            return
        for _ in range(self._num_workers):  # find the next active worker, if any
            worker_queue_idx = next(self._worker_queue_idx_cycle)
            if self._workers_status[worker_queue_idx]:
                break
        else:
            print('no worker active...')
            return

        # print(f'worker_queue_idx={worker_queue_idx}, _send_idx={self._send_idx}, index={index}')
        self._index_queues[worker_queue_idx].put((self._send_idx, index))
        self._task_info[self._send_idx] = (worker_queue_idx,)
        self._tasks_outstanding += 1
        self._send_idx += 1

    def _process_data(self, data):
        self._rcvd_idx += 1
        self._try_put_index()

        if isinstance(data, ExceptionWrapper):
            data.re_raise()

        # data = deserialize(data)

        return data

    def _mark_worker_as_unavailable(self, worker_id):
        q = self._index_queues[worker_id]
        q.put(None)  # indicate that no more data will be put on this queue by the current process

        self._workers_status[worker_id] = False

    def _shutdown_workers(self):
        # Python is shutting down
        python_exit_status = util.python_exit_status
        if python_exit_status is True or python_exit_status is None:
            return

        # shutdown all workers
        if not self._shutdown:
            self._shutdown = True

            try:
                self._workers_done_event.set()

                for worker_id in range(len(self._workers)):
                    if self._persistent_workers or self._workers_status[worker_id]:
                        self._mark_worker_as_unavailable(worker_id)

                for w in self._workers:
                    w.join(timeout=worker.MP_STATUS_CHECK_INTERVAL)

                for q in self._index_queues:
                    q.cancel_join_thread()
                    q.close()
            finally:
                if self._worker_pids_set:
                    self._worker_pids.clear()
                    self._worker_pids_set = False

                for w in self._workers:
                    if w.is_alive():
                        w.terminate()

    def __del__(self):
        self._shutdown_workers()


class DataLoader(_BaseDataLoader):
    def __init__(self, dataset: Dataset, num_workers: int = 0, **kwargs):
        """
        Args:
            dataset:
            num_workers:
            kwargs:
                - fn_to_tensor:
                - fn_worker_init:
                - transform:
                - repeat_in_batch:


        Returns:
        """
        if num_workers == 0:
            seed = tf.random.uniform((1, 1), 0, np.iinfo(np.int32).max, tf.int32).numpy()[0][0]
            set_rnd(dataset, int(seed))

        if 'fn_to_tensor' not in kwargs:
            kwargs.update({"fn_to_tensor": to_tensor.to_tf_tensor})

        if 'fn_worker_init' not in kwargs:
            kwargs.update({'fn_worker_init': self.worker_init_fn})

        if not isinstance(kwargs.get('repeat_in_batch', None), RepeatInBatch):
            logger.warning(
                f'repeat_in_batch should be instance of DatasetRepeat, but got: {kwargs.get("repeat_in_batch", None)}. '
                f'Default: RepeatInBatch("no") will be used'
            )
            kwargs.update({'repeat_in_batch': RepeatInBatch('no')})

        super().__init__(dataset=dataset, num_workers=num_workers, **kwargs)

    @staticmethod
    def worker_init_fn(worker_id: int) -> None:
        """set different random seed for the transforms in different workers

        Args:
            worker_id:

        Returns:
        """
        worker_info = worker.get_worker_info()

        set_rnd(worker_info.dataset, seed=worker_info.seed)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from dataloader import logger
from dataloader.util import get_rng


class RepeatInBatch:
    No = 'no'  # without repeat
    Append = 'append'  # append at the end, same as [x0, x1, x2] * times
    NextTo = 'next_to'  # e.g., when times = 1,  [x0, x0, x1, x1, ...]
    Shuffle = 'shuffle'  # repeat and shuffle

    kinds = {No, Append, NextTo, Shuffle}

    def __init__(self, kind, times=0):
        """

        Args
            kind: repeat kind, in ['no', 'append', 'next_to', 'shuffle']
            times: repeat times, repeat times when times > 0, otherwise no repeat
        """
        if kind not in self.kinds:
            raise ValueError(
                f'dataset repeat kind does not support: kind={kind}, supported: [{",".join(self.kinds)}]'
            )
        self.kind = kind

        if times < 0:
            raise ValueError(f'repeat times should not be less than 0. times={times}')
        self.times = times + 1

        logger.info(f'repeat data in batch with kind={self.kind}, times={times}')

        self._local_rng = get_rng(self)

    def repeat(self, batch):
        if self.kind == self.No or self.times <= 1:
            return batch

        if self.kind == self.Append:
            return batch * self.times

        if self.kind == self.NextTo:
            batch = [[data] * self.times for data in batch]
            batch = [data for mini_batch in batch for data in mini_batch]
            return batch

        if self.kind == self.Shuffle:
            batch = batch * self.times
            self._local_rng.shuffle(batch)

            return batch



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import multiprocess as multiprocessing

from dataloader.transform import apply_transform


def pool_transform(batch, transform, processes=20):
    def pool_apply_transform(elem):
        return apply_transform(transform, elem)

    pool = multiprocessing.Pool(processes=processes)

    processed = pool.map(pool_apply_transform, batch)
    processed = list(processed)

    pool.close()
    pool.join()

    return processed



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import msgpack
import msgpack_numpy

msgpack_numpy.patch()
assert msgpack.version >= (0, 5, 2)

__all__ = ['serialize', 'deserialize']


MAX_MSGPACK_LEN = 1000000000


class MsgpackSerializer:
    @staticmethod
    def serialize(obj):
        return msgpack.dumps(obj, use_bin_type=True)

    @staticmethod
    def deserialize(buf):
        return msgpack.loads(
            buf,
            raw=False,
            max_bin_len=MAX_MSGPACK_LEN,
            max_array_len=MAX_MSGPACK_LEN,
            max_map_len=MAX_MSGPACK_LEN,
            max_str_len=MAX_MSGPACK_LEN
        )


# Define the default serializer to be used that dumps data to bytes
serialize = MsgpackSerializer.serialize
deserialize = MsgpackSerializer.deserialize

# Define the default serializer to be used for passing data among a pair of peers.
# In this case the deserialization is known to happen only once
serialize_once = MsgpackSerializer.serialize
deserialize_once = MsgpackSerializer.deserialize



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import collections
import sys
import traceback
from typing import Any
from typing import Tuple

import numpy as np

from dataloader import logger

MAX_SEED = np.iinfo(np.uint32).max + 1


def _is_sequence_iterable(obj) -> bool:
    return isinstance(obj, collections.abc.Iterable) and not isinstance(obj, str)


def to_tuple(values: Any) -> Tuple[Any, ...]:
    if not _is_sequence_iterable(values):
        values = (values, )

    return tuple(values)


def bytes_to_str(text):
    if text is not None and isinstance(text, bytes):
        try:
            return text.decode('utf-8')
        except UnicodeDecodeError as e:
            logger.error(f'decode failed text={text}: {e}')
            raise e

    return text


class KeyErrorMessage(str):
    """str subclass that returns itself in repr"""
    def __repr__(self):
        return self


class ExceptionWrapper(object):
    """Wraps an exception plus traceback to communicate across threads"""
    def __init__(self, exc_info=None, where='in background'):
        if exc_info is None:
            exc_info = sys.exc_info()
        self.exc_type = exc_info[0]
        self.exc_msg = ''.join(traceback.format_exception(*exc_info))
        self.where = where

    def re_raise(self):
        r"""Re-raises the wrapped exception in the current thread"""
        msg = f'Caught {self.exc_type.__name__} {self.where}.\n Original {self.exc_msg}'

        if self.exc_type == KeyError:
            msg = KeyErrorMessage(msg)
        elif getattr(self.exc_type, 'message', None):
            raise self.exc_type(message=msg)

        try:
            exception = self.exc_type(msg)
        except TypeError:
            raise RuntimeError(msg) from None

        raise exception


def set_rnd(obj, seed):
    """

    Args:
        obj:
        seed:

    Returns:

    """
    if not hasattr(obj, '__dict__'):
        return seed

    if hasattr(obj, 'set_random_state'):
        obj.set_random_state(seed=seed % MAX_SEED)
        return seed + 1

    for key in obj.__dict__:
        if key.startswith('__'):
            continue
        seed = set_rnd(obj.__dict__[key], seed=seed)

    return seed


def get_offset(filename):
    logger.debug(f'loading offset from {filename}')
    with open(filename, 'rb') as fd:
        offset = [0]
        while fd.readline():
            offset.append(fd.tell())

        offset = offset[:-1]

    n_data = len(offset)
    logger.debug(f'loading offset done: n_offset={n_data}')

    return offset, n_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from dataloader.pipeline.datapipe import ProxyDataPipe
from dataloader.util.batch import RepeatInBatch
from dataloader.util.to_tensor import to_tf_tensor


class Batch(ProxyDataPipe):
    def __init__(self, datapipe, batch_size, drop_last, repeat_in_batch=None, to_tensor_func=to_tf_tensor):
        super().__init__(datapipe)

        self.batch_size = batch_size
        self.drop_last = drop_last

        self.repeat_in_batch = repeat_in_batch
        if repeat_in_batch is None or not isinstance(repeat_in_batch, RepeatInBatch):
            self.repeat_in_batch = RepeatInBatch(kind='no')

        self._to_tensor_func = to_tensor_func
        if self._to_tensor_func is None:
            self._to_tensor_func = to_tf_tensor

    def __len__(self):
        sz = len(self.datapipe) if self.drop_last else len(self.datapipe) + self.batch_size - 1
        return sz // self.batch_size

    def __iter__(self):
        batch = []

        for data in self.datapipe:
            batch.append(data)

            if len(batch) == self.batch_size:
                batch = self.repeat_in_batch.repeat(batch)
                yield self._to_tensor_func(batch)

                del batch[:]

        if len(batch) == self.batch_size:
            batch = self.repeat_in_batch.repeat(batch)
            yield self._to_tensor_func(batch)

        if len(batch) > 0 and not self.drop_last:
            batch = self.repeat_in_batch.repeat(batch)
            yield self._to_tensor_func(batch)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from abc import ABC
import atexit
from contextlib import contextmanager
import errno
import itertools
import multiprocessing as mp
import os
import pickle
import sys
import traceback
import uuid
import weakref

import zmq

from dataloader import logger
from dataloader.pipeline.datapipe import DataPipe
from dataloader.pipeline.datapipe import DataPipeReEntrantGuard
from dataloader.pipeline.datapipe import DataPipeTerminated
from dataloader.util.concurrency import enable_death_signal
from dataloader.util.concurrency import start_proc_mask_signal
from dataloader.util.serialize import deserialize
from dataloader.util.serialize import serialize


class _ExceptionWrapper:
    MAGIC = b"EXC_MAGIC"
    """Wraps an exception plus traceback to communicate across threads"""
    def __init__(self, exc_info):
        # It is important that we don't store exc_info, see
        # NOTE [ Python Traceback Reference Cycle Problem ]
        self.exc_type = exc_info[0]
        self.exc_msg = ''.join(traceback.format_exception(*exc_info))

    def pack(self):
        return self.MAGIC + pickle.dumps(self)

    @staticmethod
    def unpack(dp):
        if isinstance(dp, bytes) and dp.startswith(_ExceptionWrapper.MAGIC):
            return pickle.loads(dp[len(_ExceptionWrapper.MAGIC):])


def _repeat_iter(get_iter):
    while True:
        yield from get_iter()


def _bind_guard(sock, name):
    try:
        sock.bind(name)
    except zmq.ZMQError:
        logger.error(f'ZMQError in socket.bind({name})')
        raise


def _get_pipe_name(name):
    if sys.platform.startswith('linux'):
        # linux supports abstract sockets: http://api.zeromq.org/4-1:zmq-ipc
        pipe_name = "ipc://@{}-pipe-{}".format(name, str(uuid.uuid1())[:8])
        pipe_dir = os.environ.get('DL_PIPE_DIR', None)
        if pipe_dir is not None:
            logger.warning("DL_PIPE_DIR is not used on Linux any more! Abstract sockets will be used.")
    else:
        pipe_dir = os.environ.get('DL_PIPE_DIR', None)
        if pipe_dir is not None:
            logger.info("ZMQ uses DL_PIPEDIR={}".format(pipe_dir))
        else:
            pipe_dir = '.'
        assert os.path.isdir(pipe_dir), pipe_dir

        filename = '{}/{}-pipe-{}'.format(pipe_dir.rstrip('/'), name, str(uuid.uuid1())[:6])
        assert not os.path.exists(filename), "Pipe {} exists! You may be unlucky.".format(filename)

        pipe_name = "ipc://{}".format(filename)

    return pipe_name


def del_weakref(x):
    o = x()
    if o is not None:
        o.__del__()


@contextmanager
def _zmq_catch_error(name):
    try:
        yield
    except zmq.ContextTerminated:
        logger.info(f'[{name}] Context terminated')
        raise DataPipeTerminated()
    except zmq.ZMQError as e:
        if e.errno == errno.ENOTSOCK:
            logger.info(f'[{name}] Socket closed')
            raise DataPipeTerminated()

        raise
    except Exception:
        raise


class _MultiProcessZMQDataPipe(DataPipe, ABC):
    def __init__(self):
        self._reset_done = False
        self._procs = []
        self.context = None
        self.socket = None

    def reset(self):
        assert not self._reset_done, 'reset() was called twice! This violates the api'
        self._reset_done = True

        # __del__ not guaranteed to get called at exit
        atexit.register(del_weakref, weakref.ref(self))

    def _start_processes(self):
        start_proc_mask_signal(self._procs)

    def __del__(self):
        try:
            if not self._reset_done:
                return

            if not self.context.closed:
                self.socket.close(0)
                self.context.destroy(0)

            for proc in self._procs:
                proc.terminate()
                proc.join(5)
        except Exception:
            pass


class MultiProcessRunnerZMQ(_MultiProcessZMQDataPipe):
    """run a datapipe in more than 1 processes, with ZeroMQ for communication.

    It'll fork the calling process of `reset`, can collect data from the give datapipe in each process by ZMQ ipc pipe.
    """
    class Worker(mp.Process):
        def __init__(self, datapipe, conn_name, hwm, idx):
            super().__init__()

            self.datapipe = datapipe
            self.conn_name = conn_name
            self.hwm = hwm
            self.idx = idx

        def run(self):
            enable_death_signal(_warn=self.idx == 0)
            self.datapipe.reset()
            _iter = _repeat_iter(lambda: self.datapipe)

            context = zmq.Context()
            socket = context.socket(zmq.PUSH)
            socket.set_hwm(self.hwm)
            socket.connect(self.conn_name)

            try:
                while True:
                    try:
                        dp = next(_iter)
                        socket.send(serialize(dp), copy=False)
                    except Exception:
                        dp = _ExceptionWrapper(sys.exc_info()).pack()
                        socket.send(serialize(dp), copy=False)
                        raise
            except KeyboardInterrupt:
                pass
            finally:
                socket.close(0)
                context.destroy(0)

    def __init__(self, datapipe, num_procs=1, hwm=50):
        """

        Args:
            datapipe:
            num_procs: number of processes to use
            hwm: the zmq "high-water mark" (queue size) for both sender and receiver
        """
        super().__init__()

        self.datapipe = datapipe
        self.num_procs = num_procs
        self._hwm = hwm

        if num_procs > 1:
            logger.info(
                'MultiProcessRunnerZMQ will fork a datapipe more than one time, this assumes the data are i.i.d.'
            )

        try:
            self._sz = datapipe.__len__()
        except NotImplementedError:
            self._sz = -1

        self._guard = None
        self.context = None
        self.socket = None

    def reset(self):
        super().reset()
        self._guard = DataPipeReEntrantGuard()
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.PULL)
        self.socket.set_hwm(self._hwm)

        pipe_name = _get_pipe_name('datapipe')
        _bind_guard(self.socket, pipe_name)

        self._procs = [
            MultiProcessRunnerZMQ.Worker(self.datapipe, pipe_name, self._hwm, idx) for idx in range(self.num_procs)
        ]
        self._start_processes()

    def _recv(self):
        ret = deserialize(self.socket.recv(copy=False))

        exc = _ExceptionWrapper.unpack(ret)

        if exc is not None:
            logger.error(f'Exception "{str(exc.exc_type)}" in worker')
            raise exc.exc_type(exc.exc_msg)

        return ret

    def __len__(self):
        return self.datapipe.__len__()

    def __iter__(self):
        with self._guard, _zmq_catch_error('MultiProcessRunnerZMQ'):
            for k in itertools.count():
                if 0 < self._sz <= k:
                    break

                yield self._recv()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from dataloader.io.image_loader import LoadDictImage
from dataloader.io.image_reader import Cv2ImageReader
from dataloader.io.loader import Loader
from dataloader.io.reader import Reader
from dataloader.io.registry import reader_registry



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import signal
import threading

# Whether SIGCHLD handler is set for DataLoader worker failures
# Only one handler needs to be set for all DataLoaders in a process
_SIGCHLD_handler_set = False


def set_sigchld_handler():
    if not isinstance(threading.current_thread(), threading._MainThread):
        return

    global _SIGCHLD_handler_set
    if _SIGCHLD_handler_set:
        return

    previous_handler = signal.getsignal(signal.SIGCHLD)
    if not callable(previous_handler):
        previous_handler = None

    def handler(signum, frame):
        if previous_handler is not None:
            previous_handler(signum, frame)

    signal.signal(signal.SIGCHLD, handler)
    _SIGCHLD_handler_set = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import abc
from typing import Sequence
from typing import Union


class Reader:
    """Abstract class to define interface APIs to load files"""

    @abc.abstractmethod
    def read(self, data: Union[Sequence[str], str], **kwargs):
        """Read data from specified file or files

        Args:
            data: file name or a list of file names to be read
            **kwargs: additional args for actual `read` API of 3rd party libs

        Returns:
            numpy array of data
        """
        raise NotImplementedError(f'Subclass {self.__class__.__name__} must implement this method')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import atexit
from datetime import datetime
import os

_RNG_SEED = None


def fix_rng_seed(seed):
    """call at the beginning of program to fix rng seed

    Args:
        seed:

    Returns:

    """
    global _RNG_SEED
    _RNG_SEED = int(seed)


def get_rng(obj=None):
    """generate a good RNG with time, pid and the object

    Args:
        obj:

    Returns:
        np.random.RandomState: the RNG
    """
    import numpy as np

    seed = (id(obj) + os.getpid() + int(datetime.now().strftime("%Y%m%d%H%M%S%f"))) % 4294967295

    return np.random.RandomState(seed)


# interval (in seconds) to check status of processes to avoid hanging in multiprocessing
MP_STATUS_CHECK_INTERVAL = 5.0

"""Whether python is shutting down, this flag is guaranteed to be set before the python core library resources are
freed, but python may already be exiting fro some time when this is set. 

Hook to set this flag is `set_python_exit_flag`, and is inspired by a similar
hook in Python 3.7 multiprocessing library:
https://github.com/python/cpython/blob/d4d60134b29290049e28df54f23493de4f1824b6/Lib/multiprocessing/util.py#L277-L327
"""
python_exit_status = False


def set_python_exit_flag():
    global python_exit_status
    python_exit_status = True


atexit.register(set_python_exit_flag)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import logging
import os
import sys

logging_level_registry = {
    'CRITICAL': logging.CRITICAL,
    'ERROR': logging.ERROR,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG,
    None: logging.INFO
}

LOG_FMT = '[%(levelname)1.1s %(asctime)s.%(msecs)03d %(module)s:%(lineno)d] %(message)s'

logging_level = logging_level_registry[os.getenv('DL_LOGGING_LEVEL')]

logger = logging.getLogger(__name__)
logger.setLevel(logging_level)
logger.propagate = False

formatter = logging.Formatter(LOG_FMT)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
handler.setFormatter(formatter)

logger.addHandler(handler)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from dataloader.pipeline.batch import Batch
from dataloader.pipeline.dataloader import DataLoader
from dataloader.pipeline.dataset import Dataset
from dataloader.pipeline.processor import MapData
from dataloader.pipeline.processor import MapDataProcessKind
from dataloader.pipeline.processor import MultiProcessMapDataZMQ
from dataloader.pipeline.processor import MultiThreadMapData
from dataloader.pipeline.runner import MultiProcessRunnerZMQ



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from dataclasses import dataclass
import os
import queue
import random

from dataloader import logger
from dataloader.util import MP_STATUS_CHECK_INTERVAL
from dataloader.util.dataset_kind import DatasetKind
from dataloader.util.misc import ExceptionWrapper


class ManagerWatchdog:
    def __init__(self):
        self.manager_pid = os.getppid()
        self.manager_dead = False

    def is_alive(self):
        if not self.manager_dead:
            self.manager_dead = os.getppid() != self.manager_pid
        return not self.manager_dead


_worker_info = None


class WorkerInfo(object):
    __initialized = False

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.__keys = tuple(kwargs.keys())
        self.__initialized = True

    def __setattr__(self, key, val):
        if self.__initialized:
            raise RuntimeError(f'Cannot assign attributes to {self.__class__.__name__} objects')

        return super(WorkerInfo, self).__setattr__(key, val)

    def __repr__(self):
        items = []
        for k in self.__keys:
            items.append(f'{k}={getattr(self, k)}')

        return '{}({})'.format(self.__class__.__name__, ', '.join(items))


def get_worker_info():
    """

    Returns:
        current iterator worker process:
        id: the current worker id
        num_workers: the total number of workers
        seed: the random seed set for the current workers, determined by main process RNG and the worker id
        dataset: the copy of the dataset object

    """
    return _worker_info


@dataclass(frozen=True)
class IterableDatasetStopIteration(object):
    """Dummy class used to signal the end of an IterableDataset"""
    worker_id: int


@dataclass(frozen=True)
class ResumeIteration(object):
    """Dummy class used to resume the fetching when worker reuse is enabled"""
    pass


def worker_loop(
    dataset_kind,
    dataset, is_batch, drop_last, fn_to_tensor, transform, repeat_in_batch,
    index_queue, data_queue, done_event,
    base_seed, init_fn, worker_id, num_workers
):
    try:
        seed = base_seed + worker_id
        random.seed(seed)

        global _worker_info
        _worker_info = WorkerInfo(id=worker_id, num_workers=num_workers, seed=seed, dataset=dataset)

        init_exception = None

        fetcher = None

        try:
            if init_fn is not None:
                init_fn(worker_id)

            fetcher = DatasetKind.create_fetcher(
                dataset_kind, dataset, is_batch, drop_last, fn_to_tensor, transform, repeat_in_batch
            )
        except Exception as e:
            logger.error(f'init_fn or create_fetcher failed: {e} for worker: {worker_id}')
            init_exception = ExceptionWrapper(
                where=f'in DataLoader worker process {worker_id}: init_fn/create_fetcher failed'
            )

        iteration_end = False

        watchdog = ManagerWatchdog()

        while watchdog.is_alive():
            try:
                r = index_queue.get(timeout=MP_STATUS_CHECK_INTERVAL)
            except queue.Empty:
                continue

            if isinstance(r, ResumeIteration):
                # Acknowledge the main process
                data_queue.put((r, None))
                iteration_end = False
                # Recreate the fetcher for worker-reuse policy
                fetcher = DatasetKind.create_fetcher(
                    dataset_kind, dataset, is_batch, drop_last, fn_to_tensor, transform, repeat_in_batch
                )

                continue

            if r is None:
                # Received the final signal
                assert done_event.is_set() or iteration_end
                break

            if done_event.is_set() or iteration_end:
                # `done_event` is set. But I haven't received the final signal
                # (None) yet. I will keep continuing until get it, and skip the
                # processing steps.
                continue

            idx, index = r

            if init_exception is not None:
                data = init_exception
                init_exception = None
            else:
                try:
                    data = fetcher.fetch(index)
                    # data = serialize(data)
                except Exception as e:
                    if isinstance(e, StopIteration) and dataset_kind == DatasetKind.Iterable:
                        data = IterableDatasetStopIteration(worker_id)
                        # data = serialize(data)
                        # Set `iteration_end`
                        # 1. to save future `next(...)` calls, and
                        # 2. to avoid sending multiple `IterableDatasetStopIteration`s.
                        iteration_end = True
                    else:
                        logger.error(f'fetch failed index={index}, worker_id={worker_id}: {e}')
                        data = ExceptionWrapper(where="in DataLoader worker process {}".format(worker_id))

            data_queue.put((idx, data))
            del data, idx, index, r  # save memory
    except KeyboardInterrupt:
        # Main process will raise KeyboardInterrupt anyways.
        pass

    if done_event.is_set():
        data_queue.cancel_join_thread()
        data_queue.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import collections
import re

import numpy as np
import tensorflow as tf

_np_str_obj_array_pattern = re.compile(r'[SaUO]')


def to_tf_tensor(batch):
    """puts each data field into a tensor with outer dimension batch size

    Examples:
        >>> to_tf_tensor([10, 20])
        <tf.Tensor: shape=(2,), dtype=int32, numpy=array([10, 20], dtype=int32)>
    """
    elem = batch[0]
    elem_type = type(elem)

    if isinstance(elem, tf.Tensor):
        # 使用 tf.stack 和 tf.concat 在多线程加载数据时 (num_workers > 0) 会 卡住, 原因未知
        # mac python3.8 + tf2.4.1 正常, python3.7 + tf2.4.1/2.3.2 卡住
        # 线上物理机 python3.8 + tf.2.4.1 卡住
        return tf.convert_to_tensor(np.stack([b.numpy() for b in batch], axis=0))

    if elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' and elem_type.__name__ != 'string_':
        if elem_type.__name__ == 'ndarray' or elem_type.__name__ == 'memmap':
            if _np_str_obj_array_pattern.search(elem.dtype.str) is not None:
                raise TypeError(
                    f'default_collate: batch must contain tensors, numpy arrays, '
                    f'numbers, dicts or lists; found {elem.dtype}'
                )

            return to_tf_tensor([tf.convert_to_tensor(b) for b in batch])

        if elem.shape == ():  # scalars
            return tf.convert_to_tensor(batch)

    if isinstance(elem, float):
        return tf.convert_to_tensor(batch, dtype=tf.float64)

    if isinstance(elem, int):
        return tf.convert_to_tensor(batch)

    if isinstance(elem, (str, bytes)):
        return batch

    if isinstance(elem, collections.abc.Mapping):
        return {key: to_tf_tensor([d[key] for d in batch]) for key in elem}

    if isinstance(elem, tuple):
        return elem_type(to_tf_tensor(samples) for samples in zip(*batch))

    if isinstance(elem, collections.abc.Sequence):
        return tf.convert_to_tensor(batch)

    raise TypeError(
        f'default_collate: batch must contain tensors, numpy arrays, numbers, dicts or lists; found {elem_type}'
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from typing import Callable
from typing import Optional

from dataloader.pipeline.batch import Batch
from dataloader.pipeline.batch import RepeatInBatch
from dataloader.pipeline.dataset import Dataset
from dataloader.pipeline.dataset import shuffle_dataset
from dataloader.pipeline.processor import MapData
from dataloader.pipeline.processor import MapDataProcessKind
from dataloader.pipeline.processor import MultiProcessMapDataZMQ
from dataloader.pipeline.processor import MultiThreadMapData
from dataloader.pipeline.runner import MultiProcessRunnerZMQ
from dataloader.transform import apply_transform

__all__ = ['DataLoader']


class DataLoader:
    def __init__(
        self, dataset: Dataset, batch_size, shuffle=False, drop_last=False,
        repeat_in_batch_kind='no', repeat_in_batch_time=0,
        transform: Optional[Callable] = None,
        to_tensor_func: Optional[Callable] = None,
        processor_kind: MapDataProcessKind = MapDataProcessKind.NORMAL,
        num_threads=20, num_procs=20, buffer_size=10,
        runner_num_procs=1
    ):
        """load data into batch

        Args:
            dataset: @see Dataset
            batch_size:
            shuffle: shuffle data
            drop_last: drop last batch or not if len(last_batch) < batch_size, default False
            repeat_in_batch_kind: repeat kind, @see RepeatInBatch
            repeat_in_batch_time: repeat times, @see RepeatInBatch
            processor_kind: what kind of method to process the data, @see MapDataProcessKind
            num_threads: num threads used in MultiThread process
            num_procs: num processes used in MultiProcess
            buffer_size: buffer size used in MultiThread/MultiProcess
            runner_num_procs: num processes to run the process in MultiProcessRunner
            transform: callable for transforming raw data to features
            to_tensor_func: callable for mapping features to tensor, default to_tf_tensor (@see to_tensor.to_tf_tensor)
        """
        dataset = shuffle_dataset(dataset, shuffle)

        if transform is None:
            transform = lambda e: e

        if processor_kind == MapDataProcessKind.NORMAL:
            dataset = MapData(dataset, map_func=lambda e: apply_transform(transform, e))
        elif processor_kind == MapDataProcessKind.MULTI_THREAD:
            dataset = MultiThreadMapData(
                dataset, num_threads=num_threads, buffer_size=buffer_size,
                map_func=lambda e: apply_transform(transform, e)
            )
        elif processor_kind == MapDataProcessKind.MULTI_PROCESS:
            dataset = MultiProcessMapDataZMQ(
                dataset, num_procs=num_procs, buffer_size=buffer_size,
                map_func=lambda e: apply_transform(transform, e)
            )
        else:
            raise ValueError(
                f'not supported processor_kind: {processor_kind}. Choose one from [{MapDataProcessKind.kinds_str}]'
            )

        dataset = MultiProcessRunnerZMQ(dataset, num_procs=runner_num_procs)

        repeat_in_batch = RepeatInBatch(repeat_in_batch_kind, repeat_in_batch_time)
        dataset = Batch(dataset, batch_size, drop_last, repeat_in_batch, to_tensor_func)

        self.dataset = dataset
        self.dataset.reset()

    def __iter__(self):
        yield from self.dataset.__iter__()

    def __len__(self):
        return len(self.dataset)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import importlib
import types


class LazyLoader(types.ModuleType):
    """Lazily import a module, mainly to avoid pulling in large dependencies.

    refer: https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/util/lazy_loader.py
    """
    def __init__(self, local_name, parent_module_globals, name):
        self._local_name = local_name
        self._parent_module_globals = parent_module_globals

        super(LazyLoader, self).__init__(name)

    def _load(self):
        module = importlib.import_module(self.__name__)
        self._parent_module_globals[self._local_name] = module

        self.__dict__.update(module.__dict__)

        return module

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)

    def __dir__(self):
        module = self._load()
        return dir(module)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from typing import Optional
from typing import Sequence
from typing import Union

import numpy as np

from dataloader.io.reader import Reader
from dataloader.io.registry import reader_registry
from dataloader.transform import Transform


class Loader(Transform):
    """load file or files from provided path based on reader"""

    def __init__(self, reader: Union[Reader, str], *args, dtype: np.dtype = np.float32, **kwargs):
        """

        Args:
            reader: register reader to load file
            dtype: if not None convert the loaded to this data type

            *args: additional parameters for reader if providing a reader name
            **kwargs: additional parameters for reader if providing a reader name
        """
        self._reader = reader

        if isinstance(reader, str):
            reader = reader.lower()
            if reader not in reader_registry:
                raise ValueError(f'unsupported reader type: {reader}, available options: {reader_registry}')

            self._reader = reader_registry[reader](*args, **kwargs)

        self.dtype = dtype

    def __call__(self, filename: Union[Sequence[str], str], reader: Optional[Reader] = None):
        """

        Args:
            filename: path file or file-like object or a list of files
                      Will be saved in meta_data with key `filename_or_obj`
            reader:

        Returns:
            ndarray:
        """
        if reader is None:
            reader = self._reader

        if reader is None:
            raise RuntimeError(f'can not find suitable reader for this file: {filename}')

        ndarray = reader.read(filename)
        ndarray = ndarray.astype(self.dtype)

        return ndarray



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from abc import abstractmethod
from copy import copy
import multiprocessing as mp
import threading

from six.moves import queue
import zmq

from dataloader.pipeline.datapipe import DataPipeReEntrantGuard
from dataloader.pipeline.datapipe import ProxyDataPipe
from dataloader.pipeline.runner import _bind_guard
from dataloader.pipeline.runner import _get_pipe_name
from dataloader.pipeline.runner import _MultiProcessZMQDataPipe
from dataloader.pipeline.runner import _zmq_catch_error
from dataloader.util.concurrency import enable_death_signal
from dataloader.util.concurrency import StoppableThread
from dataloader.util.serialize import deserialize
from dataloader.util.serialize import serialize

__all__ = ['MultiThreadMapData', 'MultiProcessMapDataZMQ', 'MapData', 'MapDataProcessKind']


class _ParallelMapData(ProxyDataPipe):
    def __init__(self, datapipe, buffer_size):
        super().__init__(datapipe)

        if buffer_size <= 0:
            raise ValueError(f'buffer_size should be greater than 0, act: {buffer_size}')

        self._buffer_size = buffer_size
        self._buffer_occupancy = 0

        self._iter = None

    def reset(self):
        super().reset()
        dp = self.datapipe
        self._iter = dp.__iter__()

    @abstractmethod
    def _recv(self):
        pass

    @abstractmethod
    def _send(self, dp):
        pass

    def _recv_filter_none(self):
        ret = self._recv()

        if ret is None:
            raise RuntimeError(f'[{type(self)}] Map function cannot return None')

        return ret

    def _fill_buffer(self, cnt=None):
        if cnt is None:
            cnt = self._buffer_size - self._buffer_occupancy

        true_cnt = 0
        try:
            for _ in range(cnt):
                dp = next(self._iter)
                self._send(dp)
                true_cnt += 1
        except StopIteration:
            pass

        self._buffer_occupancy += true_cnt

    def get_data(self):
        self._fill_buffer()

        for dp in self._iter:
            self._send(dp)
            yield self._recv_filter_none()

        self._iter = self.datapipe.__iter__()  # refresh

        # clear buffer and re-fill
        for k in range(self._buffer_size):
            dp = self._recv_filter_none()
            self._buffer_occupancy -= 1
            if k == self._buffer_size - 1:
                self._fill_buffer()
            yield dp

    def __iter__(self):
        yield from self.get_data()


class MultiThreadMapData(_ParallelMapData):
    """
    Note:
        1. the produced data will have different ordering
        2. one should avoid starting many threads in main process to reduce GIL contention.
           The threads will only start in the process which calls :meth:`reset()`.
           Therefore you can use ``MultiProcessRunnerZMQ(MultiThreadMapData(...), 1)`` to reduce GIL contention.
    """
    class Worker(StoppableThread):
        def __init__(self, in_queue, out_queue, evt, map_func, timeout=5):
            super().__init__(evt, timeout)

            self._in_queue = in_queue
            self._out_queue = out_queue
            self.func = map_func

            self.daemon = True

        def run(self):
            try:
                while True:
                    dp = self.queue_get_stoppable(self._in_queue)
                    if self.stopped():
                        return

                    # cannot ignore None here. will lead to un-synced send/recv
                    obj = self.func(dp)
                    self.queue_put_stoppable(self._out_queue, obj)
            except Exception:
                if self.stopped():
                    pass
                else:
                    raise
            finally:
                self.stop()

    def __init__(self, datapipe, num_threads=None, map_func=None, *, buffer_size=200):
        try:
            buffer_size = min(buffer_size, len(datapipe))
        except Exception:  # datapipe may not have a len
            pass

        if num_threads is None or num_threads <= 0:
            raise ValueError('num_threads should be greater than 0')

        super().__init__(datapipe, buffer_size)

        self.num_threads = num_threads
        self.map_func = map_func
        self._threads = []
        self._evt = None
        self._in_queue = None
        self._out_queue = None
        self._guard = None

    def reset(self):
        super().reset()

        if self._threads:
            self._threads[0].stop()
            for t in self._threads:
                t.join()

        self._in_queue = queue.Queue()
        self._out_queue = queue.Queue()
        self._evt = threading.Event()

        self._threads = [
            MultiThreadMapData.Worker(self._in_queue, self._out_queue, self._evt, self.map_func)
            for _ in range(self.num_threads)
        ]

        for t in self._threads:
            t.start()

        self._guard = DataPipeReEntrantGuard()

        # Call once at the beginning, to ensure in-queue + out-queue has a total of buffer_size elements
        self._fill_buffer()

    def _recv(self):
        return self._out_queue.get()

    def _send(self, dp):
        self._in_queue.put(dp)

    def __iter__(self):
        with self._guard:
            yield from super().__iter__()

    def __del__(self):
        if self._evt is not None:
            self._evt.set()

        for p in self._threads:
            p.stop()
            p.join(timeout=5.0)


class MultiProcessMapDataZMQ(_ParallelMapData, _MultiProcessZMQDataPipe):
    """
    Note:
        1. the produced data will have different ordering
    """
    class Worker(mp.Process):
        def __init__(self, identity, map_func, pipe_name, hwm):
            super().__init__()
            self.identity = identity
            self.pipe_name = pipe_name
            self.map_func = map_func
            self.hwm = hwm

        def run(self):
            enable_death_signal(_warn=self.identity == b'0')
            ctx = zmq.Context()
            socket = ctx.socket(zmq.REP)
            socket.setsockopt(zmq.IDENTITY, self.identity)
            socket.set_hwm(self.hwm)
            socket.connect(self.pipe_name)

            while True:
                dp = deserialize(socket.recv(copy=False))
                dp = self.map_func(dp)
                socket.send(serialize(dp), copy=False)

    def __init__(self, datapipe, num_procs=None, map_func=None, *, buffer_size=200):
        try:
            buffer_size = min(buffer_size, len(datapipe))
        except Exception:  # ds may not have a length
            pass

        if num_procs is None or num_procs <= 0:
            raise ValueError('num_proc should be greater than 0')

        _ParallelMapData.__init__(self, datapipe, buffer_size)
        _MultiProcessZMQDataPipe.__init__(self)

        self.num_procs = num_procs
        self.map_func = map_func
        self._procs = []

        self._guard = None

    def reset(self):
        _MultiProcessZMQDataPipe.reset(self)
        _ParallelMapData.reset(self)

        self._guard = DataPipeReEntrantGuard()

        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.DEALER)
        self.socket.set_hwm(self._buffer_size * 2)

        pipe_name = _get_pipe_name('datapipe-map')
        _bind_guard(self.socket, pipe_name)

        worker_hwm = int(self._buffer_size * 2 // self.num_procs)
        self._procs = [
            MultiProcessMapDataZMQ.Worker(f'{i}'.encode('utf-8'), self.map_func, pipe_name, worker_hwm)
            for i in range(self.num_procs)
        ]

        self._start_processes()
        self._fill_buffer()

    def _send(self, dp):
        msg = [b'', serialize(dp)]
        self.socket.send_multipart(msg, copy=False)

    def _recv(self):
        msg = self.socket.recv_multipart(copy=False)
        dp = deserialize(msg[1])

        return dp

    def __iter__(self):
        with self._guard, _zmq_catch_error(type(self).__name__):
            yield from super().__iter__()


class MapData(ProxyDataPipe):
    """
        Note:
            1. the produced data will keep the order,
               but could be slow compare with MultiThreadMapData and MultiProcessMapDataZMQ
        """
    def __init__(self, datapipe, map_func):
        super().__init__(datapipe)
        self.map_func = map_func

    def __iter__(self):
        for dp in self.datapipe:
            ret = self.map_func(copy(dp))

            if ret is not None:
                yield ret


class MapDataProcessKind:
    NORMAL = 'normal'
    MULTI_THREAD = 'multi-thread'
    MULTI_PROCESS = 'multi-process'

    kinds = {NORMAL, MULTI_THREAD, MULTI_PROCESS}
    kinds_str = ', '.join(kinds)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from dataloader.util.fetcher import IterableDatasetFetcher
from dataloader.util.fetcher import MapDatasetFetcher


class DatasetKind:
    Map = 0
    Iterable = 1

    @staticmethod
    def create_fetcher(kind, dataset, is_batch, drop_last, fn_to_tensor, transform, repeat_in_batch):
        """get data

        Args:
            kind: Map or Iterable
            dataset:
            is_batch: sample index is  mini-batch or a single index
            drop_last: drop last batch if batch_size does not match
            fn_to_tensor: how to transform data into tensor
            transform: how to transform raw data into feature
            repeat_in_batch: repeat data item in batch

        Returns:
            Fetcher
        """
        if kind == DatasetKind.Map:
            return MapDatasetFetcher(dataset, is_batch, drop_last, fn_to_tensor, transform, repeat_in_batch)

        return IterableDatasetFetcher(dataset, is_batch, drop_last, fn_to_tensor, transform, repeat_in_batch)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from contextlib import contextmanager
import multiprocessing as mp
import platform
import signal
import threading

from six.moves import queue

from dataloader import logger


class StoppableThread(threading.Thread):
    """A thread that has a 'stop' event"""

    def __init__(self, evt=None, timeout=5):
        """

        Args:
            evt: threading.Event, if None, will create one
            timeout: put/get timeout
        """
        super().__init__()

        if evt is None:
            evt = threading.Event()

        self._stop_evt = evt
        self._timeout = timeout

    def stop(self):
        self._stop_evt.set()

    def stopped(self):
        """
        Returns:
            whether the thread is stopped or not
        """
        return self._stop_evt.isSet()

    def queue_put_stoppable(self, q, obj):
        """put obj to queue, but will give up when the thread is stopped"""
        while not self.stopped():
            try:
                q.put(obj, timeout=self._timeout)
                break
            except queue.Full:
                pass

    def queue_get_stoppable(self, q):
        """get obj from queue, but will give up when the thread is stopped"""
        while not self.stopped():
            try:
                return q.get(timeout=self._timeout)
            except queue.Empty:
                pass


def enable_death_signal(_warn=True):
    """set death signal of the current process, so the current process will be cleaned with guarantee in case the
    parent dies accidentally
    """

    if platform.system() != 'Linux':
        return

    try:
        import prctl  # pip install python-prctl
    except ImportError:
        if _warn:
            logger.warning(
                '"import prctl" failed! Install python-prctl so that processes can be cleaned with guarantee'
            )
        return
    else:
        if not hasattr(prctl, 'set_pdeathsig'):
            raise RuntimeError(
                'prctl.set_pdeathsig does not exist! Note that you need to install "python-prctl" instead of "prctl"'
            )
        # is SIGHUP a good choice?
        prctl.set_pdeathsig(signal.SIGHUP)


@contextmanager
def mask_sigint():
    """
    Returns:
        If called in main thread, returns a context where ``SIGINT`` is ignored, and yield True.
        Otherwise yield False.
    """
    if threading.current_thread() == threading.main_thread():
        sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
        yield True

        signal.signal(signal.SIGINT, sigint_handler)
    else:
        yield False


def start_proc_mask_signal(proc):
    """start processes with SIGINT ignored

    Args:
        proc: multiprocessing process or processes

    Returns:

    Note:
        The signal mask is only applied when called from main thread.
    """
    if not isinstance(proc, list):
        proc = [proc]

    with mask_sigint():
        for p in proc:
            if isinstance(p, mp.Process):
                if mp.get_start_method() == 'fork':
                    msg = (
                        'Starting a process with fork method is efficient but not safe and may cause deadlock or crash,'
                        'Use forkserver or spawn method instead if you run into such issues'
                    )
                    logger.warning(msg)

            p.start()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import abc
from typing import Any
from typing import Callable
from typing import Collection
from typing import Hashable
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np

__all__ = [
    'apply_transform',
    'Transform', 'Randomize', 'ComposeTransform', 'DictTransform', 'LambdaTransform', 'LambdaDictTransform'
]

from dataloader import logger
from dataloader.util.misc import to_tuple


def apply_transform(transform: Callable, data):
    if transform is None:
        return data

    if isinstance(data, (list, tuple)):
        return [transform(item) for item in data]

    return transform(data)


class Transform:
    @abc.abstractmethod
    def __call__(self, data: Any):
        raise NotImplementedError(f'Subclass {self.__class__.__name__} must implement this method')


class Randomize:
    R: np.random.RandomState = np.random.RandomState()
    MAX_SEED = np.iinfo(np.uint32).max + 1

    def set_random_state(self, seed: Optional[int] = None, state: Optional[np.random.RandomState] = None):
        if seed is not None:
            _seed = id(seed) if not isinstance(seed, (int, np.integer)) else seed
            _seed = _seed % self.MAX_SEED
            self.R = np.random.RandomState(_seed)

            return self

        if state is not None:
            if not isinstance(state, np.random.RandomState):
                raise TypeError(f'state must be None or a np.random.RandomState but is {type(state).__name__}')
            self.R = state
            return self

        self.R = np.random.RandomState()
        return self

    @abc.abstractmethod
    def randomize(self, data: Any):
        raise NotImplementedError(f'Subclass {self.__class__.__name__} must implement this method')


class ComposeTransform(Randomize, Transform):
    """to chain a series of calls together in a sequence"""
    def __init__(self, transforms: Optional[Union[Sequence[Callable], Callable]] = None, seed: Optional[int] = None):
        if transforms is None:
            transforms = []

        self.transforms = to_tuple(transforms)
        self.set_random_state(seed)

    def set_random_state(self, seed: Optional[int] = None, state: Optional[np.random.RandomState] = None):
        super().set_random_state(seed=seed, state=state)

        for _transform in self.transforms:
            if not isinstance(_transform, Randomize):
                continue

            _transform.set_random_state(seed=self.R.randint(self.MAX_SEED, dtype='uint32'))

        return self

    def randomize(self, data: Optional[Any] = None):
        for _transform in self.transforms:
            if not isinstance(_transform, Randomize):
                continue

            try:
                _transform.randomize(data)
            except TypeError as type_error:
                name = type(_transform).__name__
                logger.warning(
                    f'Transform "{name}" in ComposeTransform not randomized\n{name}.{type_error}.', RuntimeWarning
                )

    def __call__(self, input_):
        for _transform in self.transforms:
            input_ = apply_transform(_transform, input_)

        return input_


class DictTransform(Transform):
    def __init__(self, keys: Union[Collection[Hashable], Hashable], *args, **kwargs):
        self.keys: Tuple[Hashable, ...] = to_tuple(keys)

        if not self.keys:
            raise ValueError('keys must be non empty')

        for key in self.keys:
            if not isinstance(key, Hashable):
                raise TypeError(f'keys must be one of (Hashable, Iterable[Hashable]) but is {type(keys).__name__}')

        self.args = args
        self.kwargs = kwargs

    @abc.abstractmethod
    def __call__(self, data):
        """

        Args:
            data: a python dict, data[key] is a numpy ndarray

        Returns:

        """
        raise NotImplementedError(f'Subclass {self.__class__.__name__} must implement this method')


class LambdaTransform(Transform):
    """apply user-defined lambda func as a transform

    Examples:
        >>> image = np.ones((10, 2, 2))
        >>> image.shape
        (10, 2, 2)
        >>> trans = LambdaTransform(func=lambda x: x[:4, :, :])
        >>> trans(image).shape
        (4, 2, 2)
    """
    def __init__(self, func: Optional[Callable] = None, **kwargs):
        if func is not None and not callable(func):
            raise TypeError(f'func must be None or callable but is {type(func).__name__}')

        self.func = func
        self.kwargs = kwargs

    def __call__(self, data, func: Optional[Callable] = None):
        if func is not None:
            if not callable(func):
                raise TypeError(f'func must be None or callable but is {type(func).__name__}')

            return func(data, **self.kwargs)

        if self.func is not None:
            return self.func(data, **self.kwargs)

        raise ValueError('incompatible values: func=None and self.func=None')


class LambdaDictTransform(DictTransform):
    """apply user-defined lambda func as a map transform(almost same as LambdaTransform but on dicted data)

    Examples:
        >>> data = {'image': np.random.random((10, 2, 2)), 'label': np.ones((10, 2, 2))}
        >>> trans = LambdaDictTransform(keys=['image', 'label'], func=lambda x: x[:4, :, :])
        >>> trans(data)['label'].shape
        (4, 2, 2)
    """
    def __init__(
        self,
        keys: Union[Collection[Hashable], Hashable],
        func: Union[Sequence[Callable], Callable],
        as_single=True,
        **kwargs
    ):
        """

        Args:
            keys: data fields to be transformed by `func`
            func: user-defined funcs
            as_single: treat each element in list/tuple individually, then func will be applied on each
            **kwargs: the params used in func
        """
        super().__init__(keys)

        self._as_single = as_single

        self.func = to_tuple(func)
        self.lambda_func = LambdaTransform(**kwargs)

    def __call__(self, data):
        d = dict(data)
        for key in self.keys:
            for func in self.func:
                if isinstance(d[key], (list, tuple)) and self._as_single:
                    d[key] = [self.lambda_func(x, func=func) for x in d[key]]
                elif isinstance(d[key], dict):
                    d[key] = {k: self.lambda_func(v, func=func) for k, v in d[key].items()}
                else:
                    d[key] = self.lambda_func(d[key], func=func)

        return d



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from typing import List
from typing import Sequence
from typing import Union

import numpy as np

from dataloader.io.reader import Reader
from dataloader.util.misc import to_tuple


class Cv2ImageReader(Reader):
    def __init__(self, *args, **kwargs):
        super().__init__()

        self.args = args
        self.kwargs = kwargs

    def read(self, data: Union[Sequence[str], str], **kwargs):
        """Read image data from specified file or files

        Args:
            data:
            **kwargs:

        Returns:
            ndarray: stacked image ndarray
        """
        import cv2

        images: List[np.ndarray] = list()

        filenames: Sequence[str] = to_tuple(data)

        kwargs_ = self.kwargs.copy()
        kwargs_.update(kwargs)

        for name in filenames:
            image = cv2.imread(name, **kwargs_)
            images.append(image)

        ndarray = np.stack(images, axis=0) if len(images) > 1 else images[0]

        return ndarray



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from typing import Collection
from typing import Hashable
from typing import Optional
from typing import Union

import numpy as np

from dataloader.io.loader import Loader
from dataloader.io.reader import Reader
from dataloader.transform import DictTransform


class LoadDictImage(DictTransform):
    def __init__(
        self,
        keys: Union[Collection[Hashable], Hashable],
        *args,
        reader: Optional[Union[Reader, str]] = None,
        dtype: Optional[np.dtype] = np.float32,
        **kwargs,
    ):
        """loading images with data format like {'image_1': image_path_1, 'image_2': image_path_2, ...}

        Args:
            keys:

            reader:
            dtype:

            *args: additional parameters for reader if providing a reader name
            **kwargs: additional parameters for reader if providing a reader name
        """
        super().__init__(keys)

        self._loader = Loader(reader, *args, dtype, **kwargs)

    def __call__(self, data, reader: Optional[Reader] = None):
        """

        Args:
            data:
            reader:

        Returns:
            the loaded image
        """
        res = dict(data)
        for key in self.keys:
            res[key] = self._loader(res[key], reader)

        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import math
from typing import Generic
from typing import Iterator
from typing import List
from typing import Optional
from typing import Sized
from typing import TypeVar

import numpy as np

T_co = TypeVar('T_co', covariant=True)


class Sampler(Generic[T_co]):
    """Base class for all Samplers"""

    def __init__(self, data_source: Optional[Sized]):
        self.data_source = data_source
        if data_source is not None:
            self.n_data_source = len(data_source)

    def __iter__(self) -> Iterator[T_co]:
        raise NotImplementedError


class SequentialSampler(Sampler[int]):
    """Samples elements sequentially, always in the same order

    Args:
        data_source: dataset to sample from

    Examples:
        >>> list(SequentialSampler(range(10)))
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    """
    data_source: Sized

    def __init__(self, data_source):
        super().__init__(data_source)

    def __iter__(self):
        return iter(range(len(self.data_source)))

    def __len__(self) -> int:
        return len(self.data_source)


class RandomSampler(Sampler[int]):
    """Sample elements randomly

    If without replacement, then sample from a shuffled dataset
    If with replacement, then user can specify :attr:`num_samples` to draw.

    Args:
        data_source (Dataset): dataset to sample from
        replacement (bool): samples are drawn on-demand with replacement if True, default=False
        num_samples (int): number of samples to draw, default=`len(dataset)`. This argument is supposed to be
                           specified only when `replacement` is ``True``

    Examples:
        >>> rs = RandomSampler(range(10), False)
    """
    data_source: Sized
    replacement: bool

    def __init__(self, data_source: Sized, replacement=False, num_samples=None) -> None:
        super().__init__(data_source)

        self.replacement = replacement
        self._num_samples = num_samples
        self._default_rng = np.random.default_rng()

    @property
    def num_samples(self) -> int:
        if self._num_samples is None:
            return len(self.data_source)

        return self._num_samples

    def __iter__(self):
        if self.replacement:
            yield from self._default_rng.choice(self.n_data_source, self.num_samples, self.replacement).tolist()
        else:
            yield from self._default_rng.permutation(self.n_data_source).tolist()

    def __len__(self):
        return self.num_samples


class InfiniteConstantSampler(Sampler):
    """nothing to do. 主要用于 Iterator 类型的数据, 其仅依赖于 iter, 而不需要 index, 故不用 sampler"""
    def __init__(self):
        super().__init__(None)

    def __iter__(self):
        while True:
            yield None


class DistributedSampler(Sampler[List[int]]):
    """

    ref: torch.utils.data.distributed.DistributedSampler
    """
    def __init__(self, horovod, dataset_size, shuffle=True, seed=0, drop_last=False):
        super().__init__(None)

        if horovod is None:
            raise ValueError('horovod should not be None in DistributedSampler')

        self.dataset_size = dataset_size

        self.num_replicas = horovod.size()
        self.rank = horovod.rank()
        self.epoch = 0
        self.drop_last = drop_last

        if drop_last and dataset_size % self.num_replicas != 0:
            self.num_samples = math.ceil((dataset_size - self.num_replicas) * 1.0 / self.num_replicas)
        else:
            self.num_samples = math.ceil(dataset_size * 1.0 / self.num_replicas)

        self.total_size = self.num_samples * self.num_replicas
        self.shuffle = shuffle
        self.seed = seed

    def __len__(self):
        return self.num_samples

    def __iter__(self):
        if self.shuffle:
            indices = np.random.RandomState(seed=self.seed + self.epoch).permutation(self.dataset_size).tolist()
        else:
            indices = list(range(self.dataset_size))

        if not self.drop_last:
            # add extra samples to make it evenly divisible
            padding_size = self.total_size - len(indices)
            if padding_size <= len(indices):
                indices += indices[:padding_size]
            else:
                indices += (indices * math.ceil(padding_size / len(indices)))[:padding_size]
        else:
            # remove tail of data to make it evenly divisible.
            indices = indices[:self.total_size]

        # subsample
        indices = indices[self.rank:self.total_size:self.num_replicas]

        return iter(indices)

    def set_epoch(self, epoch):
        """Sets the epoch for this sampler

        Notes:
            when `shuffle=True`, this ensures all replicas use a different random ordering for each epoch,
            otherwise, the next iteration of this sampler will yield the same ordering.
        """
        self.epoch = epoch


class BatchSampler(Sampler[List[int]]):
    """sampler to yield a mini-batch of indices

    Args:
        sampler: (Sampler or Iterable) base sampler
        batch_size: (int) size of mini-batch
        drop_last: (bool) drop the last batch if its size is less than `batch_size`

    Example:
        >>> list(BatchSampler(SequentialSampler(range(10)), batch_size=3, drop_last=False))
        [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]
        >>> list(BatchSampler(SequentialSampler(range(10)), batch_size=3, drop_last=True))
        [[0, 1, 2], [3, 4, 5], [6, 7, 8]]
    """

    def __init__(self, sampler: Sampler[int], batch_size: int, drop_last: bool) -> None:
        super().__init__(None)

        self.sampler = sampler
        self.batch_size = batch_size
        self.drop_last = drop_last

    def __iter__(self):
        batch = []

        for idx in self.sampler:
            batch.append(idx)

            if len(batch) == self.batch_size:
                yield batch

                batch = []

        if len(batch) > 0 and not self.drop_last:
            yield batch

    def __len__(self):
        sz = len(self.sampler) if self.drop_last else len(self.sampler) + self.batch_size - 1

        return sz // self.batch_size



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

import threading

from dataloader.util import get_rng

__all__ = ['DataPipe', 'ProxyDataPipe', 'RandomDataPipe', 'DataPipeTerminated', 'DataPipeReEntrantGuard']


class DataPipeTerminated(BaseException):
    pass


class DataPipeReEntrantGuard:
    """enforce non-re-entrant"""
    def __init__(self):
        self._lock = threading.Lock()

    def __enter__(self):
        self._succeed = self._lock.acquire(False)

        if not self._succeed:
            raise threading.ThreadError("this DataPipe is not re-entrant!")

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._lock.release()

        return False


class DataPipe:
    def __len__(self):
        raise NotImplementedError()

    def reset(self):
        pass


class ProxyDataPipe(DataPipe):
    def __init__(self, datapipe):
        self.datapipe = datapipe

    def reset(self):
        self.datapipe.reset()

    def __len__(self):
        return self.datapipe.__len__()

    def __iter__(self):
        return self.datapipe.__iter__()


class RandomDataPipe(DataPipe):
    def __len__(self):
        raise NotImplementedError()

    def reset(self):
        self.rng = get_rng(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from dataloader.transform import apply_transform
from dataloader.util.batch import RepeatInBatch


class IterableDatasetFetcher:
    def __init__(self, dataset, is_batch, drop_last, fn_to_tensor, transform, repeat_in_batch: RepeatInBatch):
        """

        Args:
            dataset:
            is_batch: sample index is  mini-batch or a single index
            drop_last: drop last batch if batch_size does not match
            fn_to_tensor: how to transform data into tensor
            transform: how to transform raw data into feature
            repeat_in_batch: repeat data item in batch
        """
        self.dataset_iter = iter(dataset)
        self.is_batch = is_batch
        self.drop_last = drop_last
        self.fn_to_tensor = fn_to_tensor
        self.transform = transform
        self.repeat_in_batch = repeat_in_batch

    def fetch(self, possibly_batched_index):
        if self.is_batch:
            data = []

            for _ in possibly_batched_index:
                try:
                    data.append(next(self.dataset_iter))
                except StopIteration:
                    break

            if len(data) == 0 or (self.drop_last and len(data) < len(possibly_batched_index)):
                raise StopIteration
        else:
            data = next(self.dataset_iter)

        data = apply_transform(self.transform, data)

        if self.is_batch:
            data = self.repeat_in_batch.repeat(data)

        data = self.fn_to_tensor(data)

        return data


class MapDatasetFetcher:
    def __init__(self, dataset, is_batch, drop_last, fn_to_tensor, transform, repeat_in_batch: RepeatInBatch):
        """

        Args:
            dataset:
            is_batch: sample index is  mini-batch or a single index
            drop_last: drop last batch if batch_size does not match
            fn_to_tensor: how to transform data into tensor
            transform: how to transform raw data into feature
            repeat_in_batch: repeat data item in batch
        """
        self.dataset = dataset
        self.is_batch = is_batch
        self.drop_last = drop_last
        self.fn_to_tensor = fn_to_tensor
        self.transform = transform
        self.repeat_in_batch = repeat_in_batch

    def fetch(self, possibly_batched_index):
        if self.is_batch:
            data = [self.dataset[idx] for idx in possibly_batched_index]
        else:
            data = self.dataset[possibly_batched_index]

        data = apply_transform(self.transform, data)

        if self.is_batch:
            data = self.repeat_in_batch.repeat(data)

        data = self.fn_to_tensor(data)

        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from os import path

import setuptools
from setuptools import find_packages
from setuptools import setup

version = int(setuptools.__version__.split('.')[0])
assert version > 30, 'requires setuptools > 30'

this_directory = path.abspath(path.dirname(__file__))

try:
    import pypandoc
    long_description = pypandoc.convert_file(path.join(this_directory, 'README.md'), 'rst')
except ModuleNotFoundError:
    long_description = ''


__version__ = '0.4.0'


setup(
    name='ml-dataloader',
    url='https://github.com/ericxsun/ml-dataloader.git',
    keywords='machine learning, deep learning, neural network, data processing',
    version=__version__,
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=['examples', 'tests']),
    zip_safe=False,
    install_requires=[
        'numpy>=1.14',
        'six',
        'msgpack>=0.5.2',
        'msgpack-numpy>=0.4.4.2',
        'pyzmq>=16',
        'psutil>=5',
        'multiprocess'
    ],
    extras_require={
        'all: "linux" in sys_platform': ['python-prctl'],
    },
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#universal-wheels
    options={'bdist_wheel': {'universal': '1'}}
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#

from copy import copy

from dataloader.pipeline.datapipe import RandomDataPipe
from dataloader.raw_dataset import RawDataset
from dataloader.util import get_rng
from dataloader.util.data_kind import DataKind

__all__ = ['Dataset', 'shuffle_dataset']


class Dataset(RandomDataPipe, RawDataset):
    def __init__(self, data, kind=DataKind.MEM_SEQ):
        """

        Args:
            data:
              - list of things: if kind=DataKind.MEM_SEQ
              - filename: if kind=DataKind.FILE or kind=DataKind.MMAP_FILE
            kind: @see DataKind
        """
        super().__init__(data, kind)

        self._local_rng = get_rng(self)

        self.indices = list(range(self.n_data))
        self._iter_idx = iter(self.indices)

        self._idx = -1

    def shuffle(self):
        indices = list(range(self.n_data))
        self._local_rng.shuffle(indices)
        self._iter_idx = iter(indices)

        return self

    def __iter__(self):
        return self

    def __next__(self):
        try:
            idx = next(self._iter_idx)
            return self[idx]
        except IndexError:
            raise StopIteration()


def shuffle_dataset(dataset, shuffle):
    if not shuffle:
        return dataset

    dataset = copy(dataset)
    dataset.shuffle()
    return dataset



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
import mmap
from collections import Callable
import os
from typing import Optional

import multiprocess as multiprocessing

from dataloader import logger
from dataloader.transform import apply_transform
from dataloader.util.data_kind import DataKind
from dataloader.util.misc import bytes_to_str
from dataloader.util.misc import get_offset


class RawDataset:
    data_kinds = {DataKind.FILE, DataKind.MMAP_FILE, DataKind.MEM_SEQ}

    def __init__(self, data, kind=DataKind.MEM_SEQ, transform: Optional[Callable] = None):
        """

        Args:
            data:
              - list of things: if kind=DataKind.MEM_SEQ
              - filename: if kind=DataKind.FILE or kind=DataKind.MMAP_FILE
            kind: @see DataKind
            transform: func for mapping raw data into feature

        >>> import os, mmap
        >>> fp = open('filename', 'rb', os.O_RDONLY)
        >>> mm = mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ)
        >>> fp.close()
        >>> dataset = RawDataset(mm, DataKind.MMAP_FILE, filename='filename')

        Notes:
            - speed: DataKind.MEM_SEQ > DataKind.MMAP_FILE >> DataKind.FILE
        """

        self.kind = kind
        if kind not in self.data_kinds:
            raise ValueError(f'not supported data kind: {kind}, choose one from [{",".join(self.data_kinds)}]')

        self.offset = []
        self.n_data = 0

        if self.kind == DataKind.MEM_SEQ:
            if not isinstance(data, list):
                raise ValueError(f'if kind is DataKind.MEM_SEQ, data should be a list. type(data)={type(data)}')

            self.data = data
            self.n_data = len(data)
        elif self.kind == DataKind.FILE or self.kind == DataKind.MMAP_FILE:
            if not os.path.exists(data):
                raise ValueError(f'filename does not exist: {data}')
            self.filename = data

            self.offset, self.n_data = get_offset(self.filename)

            if self.kind == DataKind.FILE:
                self.fd = open(data, 'rb', buffering=0)
                self.lock = multiprocessing.Lock()
                logger.warning('with multiprocessing.Lock lead to poor speed')
            elif self.kind == DataKind.MMAP_FILE:
                fp = os.open(self.filename, os.O_RDONLY)
                self.data = mmap.mmap(fp, 0, access=mmap.ACCESS_READ)

                logger.warning('with mmap lead to huge memory cost')

        self.transform = transform

    def __len__(self) -> int:
        return self.n_data

    def get(self, index: int):
        data = None

        if index < 0:
            index += self.n_data

        if self.kind == DataKind.MEM_SEQ:
            data = self.data[index]
        elif self.kind == DataKind.FILE:
            with self.lock:
                self.fd.seek(self.offset[index])
                line = self.fd.readline()

            try:
                data = bytes_to_str(line).strip('\n')
            except Exception as e:
                logger.error(f'decode failed: index={index}, offset={self.offset[index]}, line={line}')
                raise e
        elif self.kind == DataKind.MMAP_FILE:
            start = self.offset[index]
            if index + 1 < self.n_data:
                line = self.data[start:self.offset[index + 1]]
            else:
                line = self.data[start:]

            try:
                data = bytes_to_str(line).split('\n')[0].strip('\n')
            except Exception as e:
                logger.error(f'decode failed: index={index}, offset={self.offset[index]}, line={line}')
                raise e

        data = apply_transform(self.transform, data)

        return data

    def __getitem__(self, index: int):
        return self.get(index)
